import copy
from dataclasses import dataclass, field
from typing import Any, TypedDict

import numpy as np
from numpy._typing import NDArray
from optype.numpy import Array1D, ToFloat, ToFloat2D
from scipy.stats import norm
from typing_extensions import Literal, final, overload, override

from relife.base import FittingResults, MaximumLikelihoodOptimizer, OptimizerConfig
from relife.lifetime_model._regression import LinearCovarEffect
from relife.utils import reshape_1d_arg

__all__ = [
    "SemiParametricProportionalHazard",
    "CoxPartialLifetimeLikelihood",
    "BreslowPartialLifetimeLikelihood",
    "EfronPartialLifetimeLikelihood",
]


@dataclass
class CoxData:
    time: NDArray[np.float64]
    covar: NDArray[np.float64]
    event: NDArray[np.bool_] | None = None
    entry: NDArray[np.float64] | None = None

    ordered_event_time: NDArray[np.float64] = field(init=False, repr=False)
    event_count: NDArray[np.int64] = field(init=False, repr=False)
    risk_set: NDArray[np.bool_] = field(init=False, repr=False)
    death_set: NDArray[np.bool_] = field(init=False, repr=False)
    ordered_event_covar: NDArray[np.float64] = field(init=False, repr=False)

    def __post_init__(self):
        self.time = reshape_1d_arg(self.time)
        self.event = (
            reshape_1d_arg(self.event)
            if self.event is not None
            else np.ones_like(self.time, dtype=np.bool_)
        )
        self.entry = (
            reshape_1d_arg(self.entry)
            if self.entry is not None
            else np.zeros_like(self.time, dtype=np.float64)
        )
        sizes = [len(x) for x in (self.time, self.event, self.entry, self.covar)]

        if len(set(sizes)) != 1:
            raise ValueError(
                f""""
                All lifetime data must have the same number of values. Fields
                length are different. Got {tuple(sizes)}.
                """
            )
        (
            self.ordered_event_time,  # uncensored sorted untied times
            ordered_event_index,
            self.event_count,
        ) = np.unique(
            self.time[self.event == 1],
            return_index=True,
            return_counts=True,
        )
        # here risk_set is mask array on time
        # left truncated & right censored
        self.risk_set = np.logical_and(
            (
                np.vstack([self.entry[:, 0]] * len(self.ordered_event_time))
                < np.hstack([self.ordered_event_time[:, None]] * len(self.time))
            ),
            (
                np.hstack([self.ordered_event_time[:, None]] * len(self.time))
                <= np.vstack([self.time[:, 0]] * len(self.ordered_event_time))
            ),
        )

        self.death_set = np.vstack(
            [self.time[:, 0] * self.event[:, 0]] * len(self.ordered_event_time)
        ) == np.hstack([self.ordered_event_time[:, None]] * len(self.time))

        self.ordered_event_covar = self.covar[self.event[:, 0] == 1][
            ordered_event_index
        ]


def psi(
    covar_effect: LinearCovarEffect,
    data: CoxData,
    on: Literal["risk"] | Literal["death"] = "risk",
    order: Literal[0] | Literal[1] | Literal[2] = 0,
) -> NDArray[np.float64]:
    """Psi formula used for likelihood computations

    Args:
        on (str, optional): "risk" or "death". Defaults to "risk". If "death",
        sum is applied on death set. order (int, optional): order derivatives
        with respect to params. Defaults to 0.

    Returns:
        np.ndarray: psi formulation
        If order 0, shape [m, 1]
        If order 1, shape [m, p]
        If order 2, shape [m, p, p]
    """
    if on == "risk":
        i_set = data.risk_set
    elif on == "death":
        i_set = data.death_set

    if order == 0:
        # shape [m]
        return np.dot(i_set, covar_effect.g(data.covar))
    elif order == 1:
        # shape [m, p]
        return np.dot(i_set, data.covar * covar_effect.g(data.covar))
    elif order == 2:
        # shape [m, p, p]
        return np.tensordot(
            i_set[:, :None],
            data.covar[:, None]
            * data.covar[:, :, None]
            * np.asarray(covar_effect.g(data.covar))[:, :, None],
            axes=1,
        ).astype(np.float64)


class _BreslowBaseline:
    """
    Class for Cox non-parametric Breslow baseline
    """

    data: CoxData
    covar_effect: LinearCovarEffect

    def __init__(self, covar_effect: LinearCovarEffect, data: CoxData):
        assert data.covar.shape[-1] == covar_effect.nb_params
        self.covar_effect = covar_effect
        self.data = data

    @overload
    def chf(self, se: Literal[False], kp: bool = False) -> NDArray[np.float64]: ...
    @overload
    def chf(
        self, se: Literal[True], kp: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...
    def chf(
        self, se: bool = False, kp: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | NDArray[np.float64]:
        """
        The cumulative hazard function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline,
            the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        # TODO : comprendre le kp
        if kp:
            values = np.cumsum(
                1
                - (
                    1
                    - (
                        self.covar_effect.g(self.data.ordered_event_covar)
                        / psi(self.covar_effect, self.data)
                    )
                )
                ** (self.covar_effect.g(self.data.ordered_event_covar))
            )
        else:
            values = np.cumsum(
                self.data.event_count[:, None] / psi(self.covar_effect, self.data)
            )
        if se:
            var = np.cumsum(
                self.data.event_count[:, None] / psi(self.covar_effect, self.data) ** 2
            )
            conf_int_values = np.hstack(
                [
                    values[:, None]
                    + np.sqrt(var)[:, None] * norm.ppf(0.05 / 2, loc=0, scale=1),
                    values[:, None]
                    - np.sqrt(var)[:, None] * norm.ppf(0.05 / 2, loc=0, scale=1),
                ]
            )
            return values, conf_int_values
        else:
            return values

    @overload
    def sf(self, se: Literal[False]) -> NDArray[np.float64]: ...
    @overload
    def sf(
        self, se: Literal[True]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...
    def sf(
        self, se: bool = False
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]] | NDArray[np.float64]:
        """
        The survival function estimation

        Parameters
        ----------
        se : bool, default is False
            If true, the estimated standard errors are returned too.

        Returns
        -------
        tuple of 2 or 3 ndarrays
            A tuple containing the timeline,
            the estimated values and optionally the estimated standard errors (if se is set to true)
        """
        if se:
            chf, chf_conf_int_values = self.chf(se=True)
            return np.exp(-chf), np.exp(-chf_conf_int_values)
        else:
            return np.exp(-self.chf(se=False))


class _SF0(TypedDict):
    timeline: NDArray[np.float64]
    estimation: NDArray[np.float64]


class NotFittedError(ValueError):
    """Exception class to raise if estimator is used before fitting."""


class SemiParametricProportionalHazard:
    """
    Class for Cox, semi-parametric, Proportional Hazards, model
    """

    fitting_results: FittingResults | None
    covar_effect: LinearCovarEffect | None
    _training_data: CoxData | None
    _sf0: _SF0 | None

    def __init__(self):
        self.fitting_results = None
        self.covar_effect = None
        self._sf0 = None
        self._training_data = None

    @property
    def params(self):
        if self.covar_effect is None:
            return None
        return self.covar_effect.params

    @property
    def nb_params(self):
        if self.covar_effect is None:
            return None
        return self.covar_effect.nb_params

    def _require_fitted(
        self,
    ) -> tuple[FittingResults, LinearCovarEffect, CoxData, _SF0]:
        """
        Sort of type narrowing function to check if
        SemiParametricProportionalHazard is fitted.
        """
        if (
            self.fitting_results is None
            or self.covar_effect is None
            or self._training_data is None
            or self._sf0 is None
        ):
            raise NotFittedError("Model not fitted")

        return (
            self.fitting_results,
            self.covar_effect,
            self._training_data,
            self._sf0,
        )

    @overload
    def sf(
        self,
        covar: NDArray[np.float64],
        se: Literal[False],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]: ...
    @overload
    def sf(
        self,
        covar: NDArray[np.float64],
        se: Literal[True],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]: ...
    @overload
    def sf(
        self, covar: NDArray[np.float64], se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
    ): ...

    def sf(
        self, covar: NDArray[np.float64], se: bool = True
    ) -> (
        tuple[NDArray[np.float64], NDArray[np.float64]]
        | tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]
    ):
        """
        The survival function estimations.

        Parameters
        ----------
        covar: np.array
            array with covariates values
        se : bool, default True
            If True, the standard errors are returned in addition to timeline
            and sf values.

        Returns
        -------
        out : tuple of timeline, values, optionally se. Default is None
            A timeline, corresponding sf values and optionnaly the standard
            errors. If the estimations does not exist yet, returns None.
        """
        fitting_results, covar_effect, _, sf0 = self._require_fitted()
        if se and fitting_results.covariance_matrix is not None:
            return (
                sf0["timeline"],
                sf0["estimation"] ** covar_effect.g(covar),
                sf0["estimation"] ** covar_effect.g(covar)
                * np.sqrt(self._q1_q2_sum(covar, fitting_results.covariance_matrix)),
            )
        return (
            sf0["timeline"],
            sf0["estimation"] ** covar_effect.g(covar),
        )

    def _q1_q2_sum(
        self, covar: NDArray[np.float64], covariance_matrix: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """
        Klein and Moeschberger: Survival Analysis Techniques for Censored and
        Truncated Data (p. 284).
        """
        _, covar_effect, training_data, _ = self._require_fitted()
        psi_values = psi(covar_effect, training_data)
        psi_order_1 = psi(covar_effect, training_data, order=1)
        d_j_on_psi = training_data.event_count[:, None] / psi_values

        q3 = np.cumsum(
            ((psi_order_1 / psi_values)[None, :, :] - covar[:, None, :])
            * d_j_on_psi[None, :, :],
            axis=1,
        )  # [m: new sample for inference, t: timeline, p]
        q2 = np.squeeze(
            np.matmul(
                q3[:, :, None, :],
                np.matmul(
                    covariance_matrix[None, None, :, :],
                    q3[:, :, :, None],
                ),
            )
        )  # [m, t]
        q1 = np.cumsum(d_j_on_psi * (1 / psi_values))
        return q1 + q2

    def init_likelihood(
        self,
        time: NDArray[np.float64],
        covar: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> "CoxPartialLifetimeLikelihood|BreslowPartialLifetimeLikelihood|EfronPartialLifetimeLikelihood":
        # init covar_effect
        self.covar_effect = LinearCovarEffect(
            (None,) * np.atleast_2d(np.asarray(covar, dtype=np.float64)).shape[-1]
        )

        x0 = kwargs.get("x0", np.random.random(covar.shape[1]))
        config = OptimizerConfig(x0)
        config.scipy_minimize_options["method"] = kwargs.get("method", "trust-exact")
        config.covariance_method = kwargs.get("covariance_method", "exact")

        cox_data = CoxData(time, covar, event=event, entry=entry)
        _, event_count = np.unique(time[event == 1], return_counts=True)
        if (event_count > 3).any():  # efron
            return EfronPartialLifetimeLikelihood(self.covar_effect, cox_data, config)
        if (event_count <= 3).all() and (2 in event_count):  # breslow
            return BreslowPartialLifetimeLikelihood(self.covar_effect, cox_data, config)

        return CoxPartialLifetimeLikelihood(self.covar_effect, cox_data, config)

    def fit(
        self,
        time: NDArray[np.float64],
        covar: NDArray[np.float64],
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ):
        likelihood = self.init_likelihood(time, covar, event, entry, **kwargs)
        self.fitting_results = likelihood.optimize()
        assert self.covar_effect is not None
        self.covar_effect.params = self.fitting_results.optimal_params
        self._training_data = likelihood.data
        # currently only BreslowBaseline is used to compute sf
        baseline = _BreslowBaseline(self.covar_effect, likelihood.data)

        timeline = likelihood.data.ordered_event_time.copy()
        self._sf0 = _SF0(timeline=timeline, estimation=baseline.sf(se=False))
        return self


@final
class CoxPartialLifetimeLikelihood(
    MaximumLikelihoodOptimizer[LinearCovarEffect, CoxData]
):
    data: CoxData
    model: LinearCovarEffect
    config: OptimizerConfig

    def __init__(
        self,
        model: LinearCovarEffect,
        data: CoxData,
        config: OptimizerConfig,
    ):
        self.model = copy.deepcopy(model)
        self.data = data
        self.config = config
        if "jac" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["jac"] = self.jac_negative_log
        if "hess" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["hess"] = self.hess_negative_log

    @property
    @override
    def nb_observations(self) -> int:
        return len(self.data.time)

    @override
    def negative_log(self, params: Array1D[np.float64]) -> ToFloat:
        self.model.params = params
        return -(
            np.log(self.model.g(self.data.ordered_event_covar)).sum()
            - np.log(psi(self.model, self.data)).sum()
        )

    def jac_negative_log(self, params: Array1D[np.float64]) -> Array1D[np.float64]:
        self.model.params = params  # changes model params

        return -(
            self.data.ordered_event_covar.sum(axis=0)
            - (psi(self.model, self.data, order=1) / psi(self.model, self.data)).sum(
                axis=0
            )
        )

    def hess_negative_log(self, params: Array1D[np.float64]) -> ToFloat2D:
        self.model.params = params  # changes model params

        psi_order_0 = psi(self.model, self.data)
        psi_order_1 = psi(self.model, self.data, order=1)

        hessian_part_1 = psi(self.model, self.data, order=2) / psi_order_0[:, :, None]
        # print("hessian_part_1 [d, p, p]:", hessian_part_1.shape)

        hessian_part_2 = (psi_order_1 / psi_order_0)[:, None] * (
            psi_order_1 / psi_order_0
        )[:, :, None]
        # print("hessian_part_2 [d, p, p]:", hessian_part_2.shape)

        return hessian_part_1.sum(axis=0) - hessian_part_2.sum(axis=0)


@final
class BreslowPartialLifetimeLikelihood(
    MaximumLikelihoodOptimizer[LinearCovarEffect, CoxData]
):
    data: CoxData
    model: LinearCovarEffect
    config: OptimizerConfig
    s_j: NDArray[np.float64]

    def __init__(
        self,
        model: LinearCovarEffect,
        data: CoxData,
        config: OptimizerConfig,
    ):
        self.model = copy.deepcopy(model)
        self.data = data
        self.config = config
        if "jac" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["jac"] = self.jac_negative_log
        if "hess" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["hess"] = self.hess_negative_log

        self.s_j = np.dot(self.data.death_set, self.data.covar)

    @property
    @override
    def nb_observations(self) -> int:
        return len(self.data.time)

    @override
    def negative_log(self, params: Array1D[np.float64]) -> ToFloat:
        self.model.params = params  # changes model params

        return -(
            np.log(self.model.g(self.s_j)).sum()
            - (
                self.data.event_count[:, None] * np.log(psi(self.model, self.data))
            ).sum()
        )

    def jac_negative_log(self, params: Array1D[np.float64]) -> Array1D[np.float64]:
        self.model.params = params  # changes model params

        return -(
            self.s_j.sum(axis=0)
            - (
                self.data.event_count[:, None]
                * (psi(self.model, self.data, order=1) / psi(self.model, self.data))
            ).sum(axis=0)
        )

    def hess_negative_log(self, params: Array1D[np.float64]) -> ToFloat2D:
        self.model.params = params  # changes model params

        psi_order_0 = psi(self.model, self.data)
        psi_order_1 = psi(self.model, self.data, order=1)

        hessian_part_1 = psi(self.model, self.data, order=2) / psi_order_0[:, :, None]
        # print("hessian_part_1 [d, p, p]:", hessian_part_1.shape)

        hessian_part_2 = (psi_order_1 / psi_order_0)[:, None] * (
            psi_order_1 / psi_order_0
        )[:, :, None]
        # print("hessian_part_2 [d, p, p]:", hessian_part_2.shape)

        return (self.data.event_count[:, None, None] * hessian_part_1).sum(axis=0) - (
            self.data.event_count[:, None, None] * hessian_part_2
        ).sum(axis=0)


@final
class EfronPartialLifetimeLikelihood(
    MaximumLikelihoodOptimizer[LinearCovarEffect, CoxData]
):
    data: CoxData
    model: LinearCovarEffect
    config: OptimizerConfig
    s_j: NDArray[np.float64]
    discount_rates: NDArray[np.float64]
    discount_rates_mask: NDArray[np.bool_]
    scipy_method = "trust-exact"

    def __init__(
        self,
        model: LinearCovarEffect,
        data: CoxData,
        config: OptimizerConfig,
    ):
        self.model = copy.deepcopy(model)
        self.data = data
        self.config = config
        if "jac" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["jac"] = self.jac_negative_log
        if "hess" not in self.config.scipy_minimize_options:
            self.config.scipy_minimize_options["hess"] = self.hess_negative_log
        self.s_j = np.dot(self.data.death_set, self.data.covar)
        self.discount_rates = (
            np.vstack(
                (np.arange(self.data.event_count.max()),) * len(self.data.event_count)
            )
            / self.data.event_count[:, None]
        )
        self.discount_rates_mask = np.where(self.discount_rates < 1, True, False)

    @property
    @override
    def nb_observations(self) -> int:
        return len(self.data.time)

    def _psi_efron(
        self,
        order: Literal[0] | Literal[1] | Literal[2] = 0,
    ) -> NDArray[np.float64]:
        """Psi formula for Efron method

        Args:
            order (int, optional): order derivatives with respect to params. Defaults to 0.

        Returns:
            np.ndarray: psi formulation for Efron method
            If order 0, shape [m, max(d_j)]
            If order 1, shape [m, max(d_j), p]
            If order 2, shape [m, max(d_j), p, p]
        """

        if order == 0:
            # shape [m, max(d_j)]
            return (
                psi(self.model, self.data, order=order) * self.discount_rates_mask
                - psi(self.model, self.data, on="death", order=order)
                * self.discount_rates
                * self.discount_rates_mask
            )
        elif order == 1:
            # shape [m, max(d_j), p]
            return (
                psi(self.model, self.data, order=1)[:, None, :]
                * self.discount_rates_mask[:, :, None]
                - psi(self.model, self.data, on="death", order=1)[:, None, :]
                * (self.discount_rates * self.discount_rates_mask)[:, :, None]
            )
        elif order == 2:
            # shape [m, max(d_j), p, p]
            return (
                psi(self.model, self.data, order=2)[:, None, :]
                * self.discount_rates_mask[:, :, None, None]
                - psi(self.model, self.data, on="death", order=2)[:, None, :]
                * (self.discount_rates * self.discount_rates_mask)[:, :, None, None]
            )

    @override
    def negative_log(
        self,
        params: NDArray[np.float64],
    ) -> float:
        self.model.params = params  # changes model params

        # .sum(axis=1, keepdims=True) --> sum on alpha to d_j
        # .sum() --> sum on j
        # using where in np.log allows to avoid 0. masked elements
        m = self._psi_efron()
        neg_L = -(
            np.log(self.model.g(self.s_j)).sum()
            - np.log(m, out=np.zeros_like(m), where=(m != 0))
            .sum(axis=1, keepdims=True)
            .sum()
        )
        return neg_L

    def jac_negative_log(self, params: Array1D[np.float64]) -> Array1D[np.float64]:
        self.model.params = params  # changes model params
        # .sum(axis=1) --> sum on alpha to d_j
        # .sum(axis=0) --> sum on j
        # using where in np.divide allows to avoid 0. masked elements
        a = self._psi_efron(order=1)
        b = self._psi_efron()[:, :, None]
        return -(
            self.s_j.sum(axis=0)
            - np.divide(a, b, out=np.zeros_like(a), where=(b != 0))
            .sum(axis=1)
            .sum(axis=0)
        )

    def hess_negative_log(self, params: Array1D[np.float64]) -> ToFloat2D:
        self.model.params = params  # changes model params

        psi_order_0 = self._psi_efron()
        psi_order_1 = self._psi_efron(order=1)

        # .sum(axis=1) --> sum on alpha to d_j
        # using where in np.divide allows to avoid 0. masked elements
        a = self._psi_efron(order=2)
        b = psi_order_0[:, :, None, None]
        hessian_part_1 = np.divide(a, b, out=np.zeros_like(a), where=(b != 0)).sum(
            axis=1
        )

        # .sum(axis=1) --> sum on alpha to d_j
        # using where in np.divide allows to avoid 0. masked elements
        b = psi_order_0[:, :, None]
        hessian_part_2 = (
            np.divide(psi_order_1, b, out=np.zeros_like(psi_order_1), where=(b != 0))[
                :, :, None, :
            ]
            * (
                np.divide(
                    psi_order_1, b, out=np.zeros_like(psi_order_1), where=(b != 0)
                )
            )[:, :, :, None]
        )
        hessian_part_2 = hessian_part_2.sum(axis=1)

        return hessian_part_1.sum(axis=0) - hessian_part_2.sum(axis=0)



"""Base classes for all parametric models."""

from __future__ import annotations

import inspect
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterator
from dataclasses import dataclass, field
from itertools import chain
from typing import (
    Any,
    Generic,
    Literal,
    Self,
    TypeVar,
    TypeVarTuple,
    final,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from optype.numpy import Array1D, Array2D, ToFloat, ToFloat1D
from scipy import stats
from scipy.optimize import approx_fprime, minimize
from typing_extensions import override

__all__ = ["ParametricModel", "FrozenParametricModel", "MaximumLikelihoodOptimizer"]


@final
class _Parameters:
    """
    Dict-like tree structured of parameters.

    Every ``ParametricModel`` are composed of a ``_Parameters`` instance.
    """

    parent: Self | None
    _leaves: dict[str, _Parameters]
    _mapping: dict[str, float]
    _all_values: tuple[float, ...]
    _all_names: tuple[str, ...]

    def __init__(self, **kwargs: float | None) -> None:
        self.parent = None
        self._leaves = {}
        self._mapping = {}
        self._all_values = ()
        self._all_names = ()
        if bool(kwargs):
            self._mapping = {
                k: v if v is not None else np.nan for k, v in kwargs.items()
            }
            self.update_tree()  # update _names and _values

    @property
    def all_names(self) -> tuple[str, ...]:
        return self._all_names

    @property
    def all_values(self) -> tuple[float, ...]:
        return self._all_values

    @property
    def size(self) -> int:
        return len(self.all_values)

    def set_leaf(self, leaf_name: str, leaf: Self) -> None:
        """
        set a leaf or new leaf
        """
        if leaf_name not in self._leaves:
            leaf.parent = self
        self._leaves[leaf_name] = leaf
        self.update_tree()  # update _names and _values

    def set_all_values(self, values: tuple[float | None, ...]) -> None:
        """set values of all tree"""
        if len(values) != self.size:
            raise ValueError(f"Expected {self.size} values but got {len(values)}")
        pos = len(self._mapping.items())
        self._mapping.update(
            zip(
                self._mapping.keys(), (np.nan if v is None else v for v in values[:pos])
            )
        )
        self._all_values = tuple(np.nan if v is None else v for v in values)
        for leaf in self._leaves.values():
            leaf.set_all_values(values[pos : pos + leaf.size])
            pos += leaf.size
        if self.parent is not None:
            self.parent.update_tree()

    def __getitem__(self, name: str) -> float:
        try:
            return self._mapping[name]
        except KeyError:
            raise ValueError(f"Parameter {name} does not exist")

    def update_tree(self) -> None:
        """update names and values of current and parent nodes"""

        def items_walk(
            parameters: _Parameters,
        ) -> Iterator[tuple[tuple[str, float], ...]]:
            yield tuple(parameters._mapping.items())
            for leaf in parameters._leaves.values():
                yield tuple(chain.from_iterable(items_walk(leaf)))

        generator = chain.from_iterable(items_walk(self))
        try:
            _k, _v = zip(*generator)
            self._all_names = _k
            self._all_values = _v
            # self._allnames, self._allvalues = zip(*generator)
        except StopIteration:
            pass
        # recursively update parent
        if self.parent is not None:
            self.parent.update_tree()


class ParametricModel:
    """
    Base class of every parametric models in ReLife.
    """

    _params: _Parameters
    _baseline_models: dict[str, ParametricModel]

    def __init__(self, **kwparams: float | None) -> None:
        self._params = _Parameters(**kwparams)
        self._baseline_models = {}

    @property
    def params(self) -> NDArray[np.float64]:
        """
        Parameters values.

        Returns
        -------
        ndarray
            Parameters values

        Notes
        -----
        If parameter values are not set, they are encoded as `np.nan` value.
        """
        return np.array(self._params.all_values)

    @params.setter
    def params(self, new_params: NDArray[np.float64]) -> None:
        if new_params.ndim > 1:
            raise ValueError(
                f"Expected params values to be 1d array. Got {new_params.ndim} ndim"
            )
        self._params.set_all_values(tuple(v.item() for v in new_params))

    @property
    def params_names(self) -> tuple[str, ...]:
        """
        Parameters names.

        Returns
        -------
        list of str
            Parameters names

        Notes
        -----
        Parameters values can be requested (a.k.a. get) by their name at instance level.
        """
        return self._params.all_names

    @property
    def nb_params(self) -> int:
        """
        Number of parameters.

        Returns
        -------
        int
            Number of parameters.

        """
        return self._params.size

    def __getattr__(self, name: str) -> Any:
        if name in self.__dict__:
            return self.__dict__[name]
        if name in super().__getattribute__("_baseline_models"):
            return super().__getattribute__("_baseline_models").get(name)
        raise AttributeError(f"{type(self).__name__} has no attribute named {name}")

    @override
    def __setattr__(self, name: str, value: Any):
        # automatically add params of new baseline model
        if isinstance(value, ParametricModel):
            self._baseline_models[name] = value
            self._params.set_leaf(f"{name}.params", value._params)
        super().__setattr__(name, value)


_ParametricModel_T = TypeVar("_ParametricModel_T", bound=ParametricModel)

Ts = TypeVarTuple("Ts")


class FrozenParametricModel(ParametricModel, Generic[_ParametricModel_T, *Ts]):
    """
    Class of every frozen parametric models.

    Frozen models encapsulate additional arguments values allowing to request
    the object without giving them.
    """

    _args: tuple[*Ts]
    _unfrozen_model: _ParametricModel_T

    def __init__(self, model: _ParametricModel_T, *args: *Ts):
        super().__init__()
        if np.any(np.isnan(model.params)):
            raise ValueError(
                f"""
                Can't freeze a model with np.nan parameters. Model params is
                {model.params}
                """
            )
        self._unfrozen_model = model
        self._args = args

    @property
    def args(self) -> tuple[*Ts]:
        return self._args

    @args.setter
    def args(self, value: tuple[*Ts]) -> None:
        if len(value) != len(self._args):
            raise ValueError
        self._args = value

    def unfreeze(self) -> _ParametricModel_T:
        return self._unfrozen_model

    @override
    def __getattr__(self, key: str) -> Any:
        frozen_type = self._unfrozen_model.__class__.__name__
        if key == "fit":
            raise AttributeError("Frozen model can't be fit")
        try:
            attr = getattr(self._unfrozen_model, key)
        except AttributeError:
            raise AttributeError(f"Frozen {frozen_type} has no attribute {key}")

        def wrapper(*args: Any, **kwargs: Any):
            return attr(*(*args, *self.args), **kwargs)

        if inspect.ismethod(attr):
            return wrapper
        return attr


@overload
def is_frozen(model: FrozenParametricModel[ParametricModel, *Ts]) -> Literal[True]: ...
@overload
def is_frozen(
    model: ParametricModel | FrozenParametricModel[ParametricModel, *Ts],
) -> bool: ...
def is_frozen(
    model: ParametricModel | FrozenParametricModel[ParametricModel, *Ts],
) -> bool:
    """
    Checks if model is frozen
    """
    from relife.base import FrozenParametricModel

    return isinstance(model, FrozenParametricModel)


@dataclass
class FittingResults:
    """Fitting results of the parametric_model core."""

    nb_observations: int  #: Number of observations (samples)
    optimal_params: NDArray[np.float64]  #: Optimal parameters values
    success: bool  #: Whether or not the optimizer exited successfully.
    neg_log_likelihood: float = field(
        repr=False
    )  #: Negative log likelihood value at optimal parameters values

    covariance_matrix: Array2D[np.float64] | None = field(
        repr=False, default=None
    )  #: Covariance matrix (computed as the inverse of the Hessian matrix).

    nb_params: int = field(init=False, repr=False)  #: Number of parameters.
    aic: float = field(init=False)  #: Akaike Information Criterion.
    aicc: float = field(
        init=False
    )  #: Akaike Information Criterion with a correction for small sample sizes.
    bic: float = field(init=False)  #: Bayesian Information Criterion.
    se: NDArray[np.float64] | None = field(
        init=False, repr=False
    )  #: Standard error, square root of the diagonal of the covariance matrix
    ic: NDArray[np.float64] | None = field(init=False, repr=False)  #: 95% IC

    def __post_init__(self):
        self.nb_params = self.optimal_params.size
        self.aic = 2 * self.nb_params + 2 * self.neg_log_likelihood
        self.aicc = self.aic + 2 * self.nb_params * (self.nb_params + 1) / (
            self.nb_observations - self.nb_params - 1
        )
        self.bic = (
            np.log(self.nb_observations) * self.nb_params + 2 * self.neg_log_likelihood
        )
        self.se = None
        self.ic = None
        if self.covariance_matrix is not None:
            self.se = np.sqrt(np.diag(self.covariance_matrix))
            self.ic = self.optimal_params.reshape(-1, 1) + stats.norm.ppf(
                (0.05, 0.95)
            ) * self.se.reshape(-1, 1) / np.sqrt(self.nb_observations)  # (p, 2)

    def se_estimation_function(
        self, jac_f: NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        """Standard error estimation function.

        Parameters
        ----------
        jac_f : 1D, 2D or 3D array
            The Jacobian of a function f with respect to params.

        Returns
        -------
        1D array
            Standard error for f(params).

        References
        ----------
        .. [1] Meeker, W. Q., Escobar, L. A., & Pascual, F. G. (2022).
            Statistical methods for reliability data. John Wiley & Sons.
        """
        # [1] equation B.10 in Appendix
        # jac_f : (p,), (p, n) or (p, m, n)
        # self.var : (p, p)
        if self.covariance_matrix is not None:
            if jac_f.ndim == 1:  # jac_f : (p,)
                return np.sqrt(
                    np.einsum("i,ij,j->", jac_f, self.covariance_matrix, jac_f)
                )  # ()
            if jac_f.ndim == 2:  # jac_f : (p, n)
                return np.sqrt(
                    np.einsum("in,ij,jn->n", jac_f, self.covariance_matrix, jac_f)
                )  # (n,)
            if (
                jac_f.ndim == 3
            ):  # jac_f : (p, m, n) if regression with more than one asset
                return np.sqrt(
                    np.einsum("imn,ij,jmn->mn", jac_f, self.covariance_matrix, jac_f)
                )  # (m,n)
            raise ValueError("Invalid jac_f ndim")
        raise ValueError("Can't compute if var is None")

    @override
    def __str__(self) -> str:
        fields = {
            "fitted params": self.optimal_params,
            "AIC": self.aic,
            "AICc": self.aicc,
            "BIC": self.bic,
        }
        # Find the maximum field name length for alignment
        max_name_length = max(len(name) for name, _ in fields.items())
        lines: list[str] = []
        for name, value in fields.items():
            # Format arrays to be more compact
            if isinstance(value, np.ndarray):
                value_str = f"[{', '.join(f'{x:.6g}' for x in value)}]"
            else:
                value_str = f"{value:.6g}" if isinstance(value, float) else str(value)
            lines.append(f"{name:<{max_name_length}} : {value_str}")
        return "\n".join(lines)


M = TypeVar("M", bound=ParametricModel)
D = TypeVar("D")


@dataclass
class OptimizerConfig:
    x0: ToFloat | ToFloat1D
    scipy_minimize_options: dict[str, Any] = field(default_factory=dict)
    covariance_method: Literal["cs", "2point", "exact", False] = False


class MaximumLikelihoodOptimizer(Generic[M, D], ABC):
    """
    Abstract maximum likelihood optimizer.

    Notes
    -----
    Jacobian and hessian are not required but they can be implemented in
    concrete likelihoods. To use the jacobian or hessian implementations in the
    likelihood, pass them into `config["scipy_minimize_options"]`.

    Attributes
    ----------
    nb_observations : int
        The number of observations.
    """

    model: M
    data: D
    config: OptimizerConfig

    @property
    @abstractmethod
    def nb_observations(self) -> int: ...

    @abstractmethod
    def negative_log(self, params: Array1D[np.float64]) -> ToFloat:
        """
        Negative log likelihood.

        Parameters
        ----------
        params : 1d array
            Parameters values.

        Returns
        -------
        out : ToFloat
            Negative log likelihood value.
        """

    def optimize(self) -> FittingResults:
        """
        Search parameters values that maximize the likelihood given data.

        Returns
        -------
        out : FittingResults
            An object that encapsulates optimal parameters and fitting
            information (AIC, variance, etc.).
        """

        optimizer = minimize(
            self.negative_log,
            self.config.x0,
            **self.config.scipy_minimize_options,
        )

        fitting_results = FittingResults(
            self.nb_observations,
            np.copy(optimizer.x),
            optimizer.success,
            optimizer.fun,
        )

        if not fitting_results.success:
            warnings.warn(
                "The negative log-likelihood minimization did not exited successfully."
            )

        if self.config.covariance_method is False:
            return fitting_results

        jac = self.config.scipy_minimize_options.get("jac", None)
        hess = self.config.scipy_minimize_options.get("hess", None)
        if jac is not None and self.config.covariance_method != "exact":
            fitting_results.covariance_matrix = approx_parameters_covariance(
                jac,
                fitting_results.optimal_params,
                method=self.config.covariance_method,
            )
        if hess is not None and self.config.covariance_method == "exact":
            fitting_results.covariance_matrix = np.linalg.pinv(
                hess(fitting_results.optimal_params)
            )
        return fitting_results


def approx_parameters_covariance(
    jac_negative_log: Callable[[Array1D[np.float64]], Array1D[np.float64]],
    optimal_params: NDArray[np.float64],
    method: Literal["2point", "cs"] = "cs",
    eps: float = 1e-6,
) -> Array2D[np.float64] | None:
    size = optimal_params.size
    hess = np.empty((size, size))

    # hessian 2 point
    if method == "2point":
        for i in range(size):
            hess[i] = approx_fprime(
                optimal_params,
                lambda x: jac_negative_log(x)[i],
                eps,
            )
        return hess
    # hessian cs
    u = eps * 1j * np.eye(size)
    complex_params = optimal_params.astype(np.complex64)  # change params to complex
    for i in range(size):
        for j in range(i, size):
            hess[i, j] = np.imag(jac_negative_log(complex_params + u[i])[j]) / eps
            if i != j:
                hess[j, i] = hess[i, j]
    covariance_matrix = None
    try:
        covariance_matrix = np.linalg.pinv(hess).astype(np.float64)
    except Exception as err:
        warnings.warn(
            f"""
            Failed to compute parameters covariance due to non-invertible
            hessian matrix. Numpy pseudo-inversion algorithm returned : {err}

            You can skip parameters covariance computation by setting
            covariance_method to False. 
            """,
            stacklevel=2,
        )

    return covariance_matrix



# pyright: basic

import numpy as np

from relife.lifetime_model import EquilibriumDistribution, LeftTruncatedModel
from relife.stochastic_process import RenewalProcess
from relife.stochastic_process._sample._iterables import RenewalProcessIterable
from relife.stochastic_process._sample.tests.utils import select_from_struct


def test_age_replacement_sampling(frozen_ar_distribution):
    renewal_process = RenewalProcess(
        frozen_ar_distribution,
        first_lifetime_model=EquilibriumDistribution(frozen_ar_distribution),
    )
    ar = frozen_ar_distribution.args[0]
    nb_assets = 1 if isinstance(ar, float) else ar.shape[0]
    nb_samples = 10
    t0 = frozen_ar_distribution.ppf(0.25)
    tf = 10 * frozen_ar_distribution.ppf(0.75)

    iterable = RenewalProcessIterable(renewal_process, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # Check that all times are less than ar for each asset
    for i in range(nb_assets):
        ar_asset = ar if isinstance(ar, float) else ar[i]
        select_asset = select_from_struct(struct_array, asset_id=i)
        np.testing.assert_array_less(select_asset["time"], ar_asset + 1e-5)


def test_left_truncated_sampling(distribution, a0):
    first_lifetime_model = LeftTruncatedModel(distribution).freeze(a0)
    renewal_process = RenewalProcess(
        distribution, first_lifetime_model=first_lifetime_model
    )
    tf = 10 * distribution.ppf(0.75)
    nb_samples = 100

    iterable = RenewalProcessIterable(renewal_process, nb_samples, (0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # check first entries are a0 for each sample
    for i in range(nb_samples):
        select_sample = select_from_struct(struct_array, sample_id=i)
        first_entries = select_sample["entry"][select_sample["entry"] > 0].reshape(
            a0.shape
        )
        np.testing.assert_equal(first_entries, a0)


def test_age_replacement_regression_sampling(frozen_ar_regression):
    renewal_process = RenewalProcess(
        frozen_ar_regression,
        first_lifetime_model=EquilibriumDistribution(frozen_ar_regression),
    )
    t0 = frozen_ar_regression.ppf(0.25).min()
    tf = 10 * frozen_ar_regression.ppf(0.75).max()
    nb_samples = 10

    iterable = RenewalProcessIterable(renewal_process, nb_samples, (t0, tf))
    struct_array = np.concatenate(tuple(iterable))
    struct_array = np.sort(struct_array, order=("asset_id", "sample_id", "timeline"))

    # check all times are bounded by the age of replacement
    # add a small constant for numerical approximations
    for i in range(frozen_ar_regression.args[0].shape[0]):
        times = select_from_struct(struct_array, asset_id=i)["time"]
        np.testing.assert_array_less(
            times, frozen_ar_regression.args[0][i].item() + 1e-5
        )



# pyright: basic

import copy
from typing import Any, TypedDict

import numpy as np
from numpy.typing import NDArray

from relife.base import ParametricModel
from relife.economic import ExponentialDiscounting, Reward
from relife.lifetime_model import (
    LeftTruncatedModel,
)
from relife.stochastic_process._sample import StochasticSampleMapping
from relife.typing import AnyParametricLifetimeModel
from relife.utils import get_model_nb_assets

from ._renewal_equations import (
    delayed_renewal_equation_solver,
    renewal_equation_solver,
)


def _make_timeline(tf: float, nb_steps: int) -> NDArray[np.float64]:
    timeline = np.linspace(0, tf, nb_steps, dtype=np.float64)  # (nb_steps,)
    return np.atleast_2d(timeline)  # (1, nb_steps) to ensure broadcasting


class LifetimeFitArgs(TypedDict):
    time: NDArray[np.floating]
    event: NDArray[np.bool_]
    entry: NDArray[np.floating]
    args: tuple[NDArray[np.floating], ...]


class RenewalProcess(ParametricModel):
    # noinspection PyUnresolvedReferences
    """Renewal process.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.

    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default

    Attributes
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.

    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    nb_params
    params
    params_names
    """

    lifetime_model: AnyParametricLifetimeModel[()]
    first_lifetime_model: AnyParametricLifetimeModel[()] | None

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        first_lifetime_model: AnyParametricLifetimeModel[()] | None = None,
    ) -> None:
        super().__init__()
        self.lifetime_model = lifetime_model
        self.first_lifetime_model = first_lifetime_model

    def renewal_function(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""The renewal function.

        The renewal function gives the expected total number of renewals. It is computed  by solving the
        renewal equation:

        .. math::

            m(t) = F_1(t) + \int_0^t m(t-x) \mathrm{d}F(x)

        where:

        - :math:`m` is the renewal function.
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model.
        - :math:`F_1` is the cumulative distribution function of the underlying
          lifetime model for the fist renewal in the case of a delayed renewal
          process.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.

        References
        ----------
        .. [1] Rausand, M., Barros, A., & Hoyland, A. (2020). System Reliability
            Theory: Models, Statistical Methods, and Applications. John Wiley &
            Sons.
        """

        timeline = _make_timeline(tf, nb_steps)  # (nb_steps,) or (1, nb_steps)
        renewal_function = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            self.lifetime_model.cdf
            if self.first_lifetime_model is None
            else self.first_lifetime_model.cdf,
        )
        return np.squeeze(timeline), np.squeeze(renewal_function)

    def renewal_density(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""The renewal density.

        The renewal density corresponds to the derivative of the renewal function with
        respect to time. It is computed by solving the renewal equation:

        .. math::

            \mu(t) = f_1(t) + \int_0^t \mu(t-x) \mathrm{d}F(x)

        where:

        - :math:`\mu` is the renewal function.
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model.
        - :math:`f_1` is the probability density function of the underlying
          lifetime model for the fist renewal in the case of a delayed renewal
          process.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.

        References
        ----------
        .. [1] Rausand, M., Barros, A., & Hoyland, A. (2020). System Reliability
            Theory: Models, Statistical Methods, and Applications. John Wiley &
            Sons.
        """
        timeline = _make_timeline(tf, nb_steps)  #  (nb_steps,) or (m, nb_steps)
        renewal_density = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            self.lifetime_model.pdf
            if self.first_lifetime_model is None
            else self.first_lifetime_model.pdf,
        )
        return np.squeeze(timeline), np.squeeze(renewal_density)

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        nb_samples : int
            The size of the desired sample
        time_window : tuple of two floats
            Time window in which data are sampled
        seed : int, optional
            Random seed, by default None.

        """

        from ._sample import RenewalProcessIterable

        iterable = RenewalProcessIterable(self, nb_samples, time_window, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )
        return StochasticSampleMapping.from_struct_array(
            struct_array, get_model_nb_assets(self), nb_samples
        )

    def generate_failure_data(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ) -> dict[str, Any]:
        """Generate lifetime data

        This function will generate lifetime data that can be used to fit a lifetime model.

        Parameters
        ----------
        nb_samples : int
            The size of the desired sample
        time_window : tuple of two floats
            Time window in which data are sampled
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of time, event, entry and args (covariates)

        """
        from relife.base import FrozenParametricModel

        from ._sample import RenewalProcessIterable

        if (
            self.first_lifetime_model is not None
            and self.first_lifetime_model != self.lifetime_model
        ):
            if isinstance(self.first_lifetime_model, FrozenParametricModel):
                if (
                    isinstance(
                        self.first_lifetime_model._unfrozen_model, LeftTruncatedModel
                    )
                    and self.first_lifetime_model._unfrozen_model.baseline
                    == self.lifetime_model
                ):
                    pass
            else:
                raise ValueError(
                    "Calling sample_lifetime_data with lifetime_model different from first_lifetime_model is ambiguous."
                )
        iterable = RenewalProcessIterable(self, nb_samples, time_window, seed=seed)
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("sample_id", "asset_id", "timeline")
        )

        args_2d = tuple(
            (np.atleast_2d(arg) for arg in getattr(self.lifetime_model, "args", ()))
        )
        tuple_args_arr = tuple(
            (
                np.take(np.asarray(arg), struct_array["asset_id"], axis=0)
                for arg in args_2d
            )
        )

        returned_dict = {
            "time": struct_array["time"].copy(),
            "event": struct_array["event"].copy(),
            "entry": struct_array["entry"].copy(),
            "args": tuple_args_arr,
        }
        return returned_dict


class RenewalRewardProcess(RenewalProcess):
    # noinspection PyUnresolvedReferences
    """Renewal reward process.

    Parameters
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    reward : Reward
        A reward object that answers costs or conditional costs given lifetime values
    discounting_rate : float
        The discounting rate value used in the exponential discounting function
    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    reward : Reward
        A reward object for the first renewal

    Attributes
    ----------
    lifetime_model : any lifetime distribution or frozen lifetime model
        A lifetime model representing the durations between events.
    first_lifetime_model : any lifetime distribution or frozen lifetime model, optional
        A lifetime model for the first renewal (delayed renewal process). It is None by default
    reward : Reward
        A reward object that answers costs or conditional costs given lifetime values
    first_reward : Reward
        A reward object for the first renewal. If it is not given at the initialization, it is a copy of reward.
    discounting_rate
    nb_params
    params
    params_names
    """

    lifetime_model: AnyParametricLifetimeModel[()]
    first_lifetime_model: AnyParametricLifetimeModel[()] | None
    reward: Reward
    first_reward: Reward
    discounting: ExponentialDiscounting

    def __init__(
        self,
        lifetime_model: AnyParametricLifetimeModel[()],
        reward: Reward,
        discounting_rate: float = 0.0,
        first_lifetime_model: AnyParametricLifetimeModel[()] | None = None,
        first_reward: Reward | None = None,
    ) -> None:
        super().__init__(lifetime_model, first_lifetime_model)
        self.reward = reward
        self.first_reward = (
            first_reward if first_reward is not None else copy.deepcopy(reward)
        )
        self.discounting = ExponentialDiscounting(discounting_rate)

    @property
    def discounting_rate(self) -> float:
        """
        The discounting rate value
        """
        return self.discounting.rate

    @discounting_rate.setter
    def discounting_rate(self, value: float) -> None:
        self.discounting.rate = value

    def expected_total_reward(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        r"""The expected total reward.

        The renewal equation solved to compute the expected reward is:

        .. math::

            z(t) = \int_0^t E[Y | X = x] e^{-\delta x} \mathrm{d}F(x) + \int_0^t z(t-x)
            e^{-\delta x}\mathrm{d}F(x)

        where:

        - :math:`z` is the expected total reward.
        - :math:`F` is the cumulative distribution function of the underlying
          lifetime model.
        - :math:`X` the interarrival random variable.
        - :math:`Y` the associated reward.
        - :math:`D` the exponential discount factor.

        If the renewal reward process is delayed, the expected total reward is
        modified as:

        .. math::

            z_1(t) = \int_0^t E[Y_1 | X_1 = x] e^{-\delta x} \mathrm{d}F_1(x) + \int_0^t
            z(t-x) e^{-\delta x} \mathrm{d}F_1(x)

        where:

        - :math:`z_1` is the expected total reward with delay.
        - :math:`F_1` is the cumulative distribution function of the lifetime
          model for the first renewal.
        - :math:`X_1` the interarrival random variable of the first renewal.
        - :math:`Y_1` the associated reward of the first renewal.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.

        """
        timeline = _make_timeline(tf, nb_steps)  #  (nb_steps,) or (m, nb_steps)
        z = renewal_equation_solver(
            timeline,
            self.lifetime_model,
            lambda t: self.lifetime_model.ls_integrate(
                lambda x: (
                    self.reward.conditional_expectation(x) * self.discounting.factor(x)
                ),
                np.zeros_like(t),
                np.asarray(t),
                deg=15,
            ),  # reward partial expectation
            discounting=self.discounting,
        )
        if self.first_lifetime_model is not None:
            z = delayed_renewal_equation_solver(
                timeline,
                z,
                self.first_lifetime_model,
                lambda t: self.first_lifetime_model.ls_integrate(
                    lambda x: (
                        self.first_reward.conditional_expectation(x)
                        * self.discounting.factor(x)
                    ),
                    np.zeros_like(t),
                    np.asarray(t),
                    deg=15,
                ),  # reward partial expectation
                discounting=self.discounting,
            )
        return np.squeeze(timeline), np.squeeze(
            z
        )  # (nb_steps,), (nb_steps,) or (m, nb_steps)

    def asymptotic_expected_total_reward(self) -> np.float64 | NDArray[np.float64]:
        r"""Asymptotic expected total reward.

        The asymptotic expected total reward is:

        .. math::

            z^\infty = \lim_{t\to \infty} z(t) = \dfrac{E\left[Y e^{-\delta X}\right]}{1-E\left[e^{-\delta X}\right]}

        where:

        - :math:`X` the interarrival random variable.
        - :math:`Y` the associated reward.
        - :math:`D` the exponential discount factor.

        If the renewal reward process is delayed, the asymptotic expected total
        reward is modified as:

        .. math::

            z_1^\infty = E\left[Y_1 e^{-\delta X_1}\right] + z^\infty E\left[e^{-\delta X_1}\right]

        where:

        - :math:`X_1` the interarrival random variable of the first renewal.
        - :math:`Y_1` the associated reward of the first renewal.

        Returns
        -------
        ndarray
            The assymptotic expected total reward of the process.
        """
        lf = self.lifetime_model.ls_integrate(
            lambda x: self.discounting.factor(x),
            np.float64(0.0),
            np.asarray(np.inf),
            deg=100,
        )  # () or (m, 1)
        if self.discounting_rate == 0.0:
            return np.full_like(np.squeeze(lf), np.inf)
        ly = self.lifetime_model.ls_integrate(
            lambda x: (
                self.discounting.factor(x) * self.reward.conditional_expectation(x)
            ),
            0.0,
            np.inf,
            deg=100,
        )  # () or (m, 1)
        z = np.squeeze(ly / (1 - lf))  # () or (m,)
        if self.first_lifetime_model is not None:
            lf1 = np.squeeze(
                self.first_lifetime_model.ls_integrate(
                    lambda x: self.discounting.factor(x), 0.0, np.inf, deg=100
                )
            )  # () or (m,)
            ly1 = np.squeeze(
                self.first_lifetime_model.ls_integrate(
                    lambda x: (
                        self.discounting.factor(x)
                        * self.first_reward.conditional_expectation(x)
                    ),
                    0.0,
                    np.inf,
                    deg=100,
                )
            )  # () or (m,)
            z = ly1 + z * lf1  # () or (m,)
        return z  # () or (m,)

    def expected_equivalent_annual_worth(
        self, tf: float, nb_steps: int
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Expected equivalent annual worth.

        Gives the equivalent annual worth of the expected total reward of the
        process at each point of the timeline.

        The equivalent annual worth at time :math:`t` is equal to the expected
        total reward :math:`z` divided by the annuity factor :math:`AF(t)`.

        Parameters
        ----------
        tf : float
            The final time.
        nb_steps : int
            The number of steps used to discretized the time.

        Returns
        -------
        tuple of two ndarrays
            A tuple containing the timeline and the computed values.
        """
        timeline, z = self.expected_total_reward(
            tf, nb_steps
        )  # timeline : (nb_steps,) z : (nb_steps,) or (m, nb_steps)
        af = self.discounting.annuity_factor(timeline)  # (nb_steps,)
        if z.ndim == 2 and af.shape != z.shape:  # (m, nb_steps)
            af = np.tile(af, (z.shape[0], 1))  # (m, nb_steps)
        q = z / (af + 1e-6)  # # (nb_steps,) or (m, nb_steps) avoid zero division
        if self.first_lifetime_model is not None:
            q0 = self.first_reward.conditional_expectation(
                np.asarray(0.0)
            ) * self.first_lifetime_model.pdf(0.0)
        else:
            q0 = self.reward.conditional_expectation(
                np.asarray(0.0)
            ) * self.lifetime_model.pdf(0.0)
        # q0 : () or (m, 1)
        q0 = np.broadcast_to(q0, af.shape)  # (), (nb_steps,) or (m, nb_steps)
        eeac = np.where(af == 0, q0, q)  # (nb_steps,) or (m, nb_steps)
        return np.squeeze(timeline), np.squeeze(
            eeac
        )  # (nb_steps,) and (nb_steps) or (m, nb_steps)

    def asymptotic_expected_equivalent_annual_worth(
        self,
    ) -> np.float64 | NDArray[np.float64]:
        """Asymptotic expected equivalent annual worth.

        Returns
        -------
        ndarray
            The assymptotic expected equivalent annual worth.
        """
        if self.discounting_rate == 0.0:
            return np.squeeze(
                np.asarray(
                    self.lifetime_model.ls_integrate(
                        lambda x: self.reward.conditional_expectation(x),
                        np.float64(0.0),
                        np.asarray(np.inf),
                        deg=100,
                    ),
                    dtype=float,
                )
                / self.lifetime_model.mean()
            )  # () or (m,)
        return (
            self.discounting_rate * self.asymptotic_expected_total_reward()
        )  # () or (m,)



from ._preventive_age_replacement import (
    AgeReplacementPolicy,
    NonHomogeneousPoissonAgeReplacementPolicy,
    OneCycleAgeReplacementPolicy,
    age_replacement_policy,
)
from ._run_to_failure import (
    OneCycleRunToFailurePolicy,
    RunToFailurePolicy,
    run_to_failure_policy,
)

__all__ = [
    "age_replacement_policy",
    "run_to_failure_policy",
    "AgeReplacementPolicy",
    "NonHomogeneousPoissonAgeReplacementPolicy",
    "OneCycleAgeReplacementPolicy",
    "OneCycleRunToFailurePolicy",
    "RunToFailurePolicy",
]



# pyright: basic

from __future__ import annotations

from typing import Any, Generic, Self, TypeVarTuple

import numpy as np

from relife.base import FrozenParametricModel, ParametricModel, FittingResults
from relife.lifetime_model._base import FittableParametricLifetimeModel
from relife.stochastic_process._sample import StochasticSampleMapping

Ts = TypeVarTuple("Ts")

__all__ = ["Kijima1Process", "Kijima2Process"]


class Kijima1Process(ParametricModel, Generic[*Ts]):
    """
    Kijima I Process.
    """

    lifetime_model: FittableParametricLifetimeModel[*Ts]
    fitting_results: FittingResults | None

    def __init__(
        self,
        lifetime_model: FittableParametricLifetimeModel[*Ts],
        q: float | None = None,
    ):
        super().__init__(q=q)
        self.lifetime_model = lifetime_model
        self.fitting_results = None

    @property
    def q(self):
        return self.params[0]

    def freeze(self, *args: *Ts) -> FrozenKijima1Process[*Ts]:
        """
        Freeze any arguments required by the process into the object data.

        Parameters
        ----------
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenParametricModel
        """
        return FrozenKijima1Process(self, *args)

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], *args, seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in a StochasticSampleMapping object.

        Parameters
        ----------
        nb_samples : int
            The number of samples
        time_window : tuple of two floats
            Time window in which data are sampled
        *args : float or np.ndarray
            Additional arguments needed by the model.
        seed : int, optional
            Random seed, by default None.

        """
        frozen_kijima = self.freeze(*args)
        return frozen_kijima.sample(
            nb_samples=nb_samples, time_window=time_window, seed=None
        )

    def generate_failure_data(self) -> dict[str, Any]:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Failure data methods for stochastic processes will be introduced in a future release"
        )

    def fit(self) -> Self:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Fitting methods for Kijima processes will be introduced in a future release"
        )


class FrozenKijima1Process(FrozenParametricModel[Kijima1Process[*Ts], *Ts]):
    """
    Kijima I process.
    """

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        """

        from relife.utils import get_model_nb_assets

        from ._sample import Kijima1ProcessIterable

        iterable = Kijima1ProcessIterable(
            self, nb_samples, time_window=time_window, seed=seed
        )
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )
        return StochasticSampleMapping.from_struct_array(
            struct_array, get_model_nb_assets(self), nb_samples
        )

    def generate_failure_data(self) -> dict[str, Any]:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Failure data methods for stochastic processes will be introduced in a future release"
        )


class Kijima2Process(ParametricModel, Generic[*Ts]):
    """
    Kijima II Process.
    """

    lifetime_model: FittableParametricLifetimeModel[*Ts]
    fitting_results: FittingResults | None

    def __init__(
        self,
        lifetime_model: FittableParametricLifetimeModel[*Ts],
        q: float | None = None,
    ):
        super().__init__(q=q)
        self.lifetime_model = lifetime_model
        self.fitting_results = None

    @property
    def q(self):
        return self.params[0]

    def freeze(self, *args: *Ts) -> FrozenKijima2Process[*Ts]:
        """
        Freeze any arguments required by the process into the object data.

        Parameters
        ----------
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenParametricModel
        """
        return FrozenKijima2Process(self, *args)

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], *args, seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in a StochasticSampleMapping object.

        Parameters
        ----------
        nb_samples : int
            The number of samples
        time_window : tuple of two floats
            Time window in which data are sampled
        *args : float or np.ndarray
            Additional arguments needed by the model.
        seed : int, optional
            Random seed, by default None.

        """
        frozen_kijima = self.freeze(*args)
        return frozen_kijima.sample(
            nb_samples=nb_samples, time_window=time_window, seed=None
        )

    def generate_failure_data(self) -> dict[str, Any]:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Failure data methods for stochastic processes will be introduced in a future release"
        )

    def fit(self) -> Self:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Fitting methods for Kijima processes will be introduced in a future release"
        )


class FrozenKijima2Process(FrozenParametricModel[Kijima2Process[*Ts], *Ts]):
    """
    Kijima II process.
    """

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        """

        from relife.utils import get_model_nb_assets

        from ._sample import Kijima2ProcessIterable

        iterable = Kijima2ProcessIterable(
            self, nb_samples, time_window=time_window, seed=seed
        )
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )
        return StochasticSampleMapping.from_struct_array(
            struct_array, get_model_nb_assets(self), nb_samples
        )

    def generate_failure_data(self) -> dict[str, Any]:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Failure data methods for stochastic processes will be introduced in a future release"
        )



"""Lifetime distributions."""

from __future__ import annotations

from abc import ABC
from typing import (
    Any,
    Callable,
    Literal,
    Self,
    TypeAlias,
    TypeVar,
    TypeVarTuple,
    final,
    overload,
)

import numpy as np
import numpydoc.docscrape as docscrape  # pyright: ignore[reportMissingTypeStubs]
from numpy.typing import NDArray
from scipy.optimize import Bounds, newton
from scipy.special import digamma, exp1, gamma, gammaincc, gammainccinv
from typing_extensions import override

from relife.base import OptimizerConfig
from relife.typing import (
    AnyFloat,
    NumpyBool,
    NumpyFloat,
    Seed,
)
from relife.utils.quadrature import laguerre_quadrature, legendre_quadrature

from ._base import (
    FittableParametricLifetimeModel,
    LifetimeData,
    LifetimeLikelihood,
    ParametricLifetimeModel,
    document_args,
)

__all__: list[str] = [
    "Gompertz",
    "Weibull",
    "Gamma",
    "LogLogistic",
    "EquilibriumDistribution",
    "Exponential",
    "MinimumDistribution",
]


M = TypeVar(
    "M",
    bound=FittableParametricLifetimeModel[()],
)


class LifetimeDistribution(FittableParametricLifetimeModel[()], ABC):
    """
    Base class for distribution model.
    """

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def sf(self, time: AnyFloat) -> NumpyFloat:
        return super().sf(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def isf(self, probability: AnyFloat) -> NumpyFloat:
        cumulative_hazard_rate = -np.log(
            np.clip(probability, 0, 1 - np.finfo(float).resolution)
        )
        return self.ichf(cumulative_hazard_rate)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def cdf(self, time: AnyFloat) -> NumpyFloat:
        return super().cdf(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def pdf(self, time: AnyFloat) -> NumpyFloat:
        return super().pdf(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def ppf(self, probability: AnyFloat) -> NumpyFloat:
        return super().ppf(probability)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def moment(self, n: int) -> NumpyFloat:
        return super().moment(n)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def median(self) -> NumpyFloat:
        return self.ppf(0.5)  # no super here to return np.float64

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def jac_sf(self, time: AnyFloat) -> NumpyFloat:
        jac_chf, sf = self.jac_chf(time), self.sf(time)
        return -jac_chf * sf

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def jac_cdf(self, time: AnyFloat) -> NumpyFloat:
        return super().jac_cdf(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def jac_pdf(self, time: AnyFloat) -> NumpyFloat:
        jac_hf, hf = self.jac_hf(time), self.hf(time)
        jac_sf, sf = self.jac_sf(time), self.sf(time)
        return jac_hf * sf + jac_sf * hf

    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> NumpyFloat: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Seed | None = None,
    ) -> tuple[NumpyFloat, NumpyBool, NumpyFloat]: ...
    @overload
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ): ...
    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def rvs(
        self,
        size: int | tuple[int, int],
        *,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Seed | None = None,
    ) -> (
        NumpyFloat
        | tuple[NumpyFloat, NumpyBool]
        | tuple[NumpyFloat, NumpyFloat]
        | tuple[NumpyFloat, NumpyBool, NumpyFloat]
    ):
        return super().rvs(
            size,
            return_event=return_event,
            return_entry=return_entry,
            seed=seed,
        )

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        *,
        deg: int = 10,
    ) -> NumpyFloat:
        return super().ls_integrate(func, a, b, deg=deg)

    @override
    def init_likelihood(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        **kwargs: Any,
    ) -> LifetimeLikelihood[Self]:
        assert args is None
        lifetime_data = LifetimeData(time, event=event, entry=entry)
        x0 = kwargs.get("x0", init_distrib_params_from_lifetimes(self, lifetime_data))
        config = OptimizerConfig(x0)
        config.scipy_minimize_options["bounds"] = kwargs.get(
            "bounds", get_distrib_params_bounds(self)
        )
        config.scipy_minimize_options["method"] = kwargs.get("method", "L-BFGS-B")
        config.covariance_method = kwargs.get(
            "covariance_method", "2point" if isinstance(self, Gamma) else "cs"
        )
        optimizer = LifetimeLikelihood(self, lifetime_data, config)
        return optimizer


def init_distrib_params_from_lifetimes(
    model: LifetimeDistribution, data: LifetimeData
) -> NDArray[np.float64]:
    # flatten censored_time in case it is 2D
    all_time_values = np.concatenate(
        (data.complete_time.flatten(), data.censored_time.flatten())
    )
    if isinstance(model, Gompertz):
        param0 = np.empty(model.nb_params, dtype=np.float64)
        rate = np.pi / (np.sqrt(6) * np.std(all_time_values))
        shape = np.exp(-rate * np.mean(all_time_values))
        param0[0] = shape
        param0[1] = rate
        return param0

    param0 = np.ones(model.nb_params, dtype=np.float64)
    param0[-1] = 1 / np.median(all_time_values)
    return param0


def get_distrib_params_bounds(model: LifetimeDistribution) -> Bounds:
    return Bounds(
        np.full(model.nb_params, np.finfo(float).resolution),
        np.full(model.nb_params, np.inf),
    )


@final
class Exponential(LifetimeDistribution):
    r"""
    Exponential lifetime distribution.

    The exponential distribution is a 1-parameter distribution with
    :math:`(\lambda)`. The probability density function is:

    .. math::

        f(t) = \lambda e^{-\lambda t}

    where:
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    rate : float, default is None
        Rate parameter.

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    rate
    """

    def __init__(self, rate: float | None = None):
        super().__init__(rate=rate)

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the rate value.

        Returns
        -------
        out: float
        """
        return self._params["rate"]

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def hf(self, time: AnyFloat) -> NumpyFloat:
        return self.rate * np.ones_like(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def chf(self, time: AnyFloat) -> NumpyFloat:
        return np.asarray(self.rate) * time

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def mean(self) -> NumpyFloat:
        return 1 / np.float64(self.rate)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def var(self) -> NumpyFloat:
        return 1 / np.float64(self.rate) ** 2

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def mrl(self, time: AnyFloat) -> NumpyFloat:
        return 1 / self.rate * np.ones_like(time)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return cumulative_hazard_rate / np.asarray(self.rate)

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def jac_hf(self, time: AnyFloat) -> NumpyFloat:
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(np.ones_like(time, dtype=np.float64), axis=0).copy()
        else:
            jac = np.array([1], dtype=np.float64)
        return jac

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def jac_chf(self, time: AnyFloat) -> NumpyFloat:
        if isinstance(time, np.ndarray):
            jac = np.expand_dims(time, axis=0).copy().astype(np.float64)
        else:
            jac = np.array([time], dtype=np.float64)
        return jac

    @override
    @document_args(base_cls=FittableParametricLifetimeModel, args_docstring=[])
    def dhf(self, time: AnyFloat) -> NumpyFloat:
        if isinstance(time, np.ndarray):
            return np.zeros_like(time, dtype=np.float64)
        return np.asarray(0.0)


@final
class Weibull(LifetimeDistribution):
    r"""
    Weibull lifetime distribution.

    The Weibull distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda (\lambda t)^{c-1} e^{-(\lambda t)^c}

    where:
        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    Parameters
    ----------
    shape : float, default is None
        Shape parameter.
    rate : float, default is None
        Rate parameter.

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: float | None = None, rate: float | None = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """
        Returns shape value.

        Returns
        -------
        out: float
        """
        return self._params["shape"]

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the rate value.

        Returns
        -------
        out: float
        """
        return self._params["rate"]

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def hf(self, time: AnyFloat) -> NumpyFloat:
        return (
            self.shape * self.rate * (self.rate * np.asarray(time)) ** (self.shape - 1)
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def chf(self, time: AnyFloat) -> NumpyFloat:
        return (self.rate * np.asarray(time)) ** self.shape

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def mean(self) -> NumpyFloat:
        return gamma(1 + 1 / self.shape) / self.rate

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def var(self) -> NumpyFloat:
        return gamma(1 + 2 / self.shape) / self.rate**2 - self.mean() ** 2

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def mrl(self, time: AnyFloat) -> NumpyFloat:
        return (
            gamma(1 / self.shape)
            / (self.rate * self.shape * self.sf(time))
            * gammaincc(
                1 / self.shape,
                (self.rate * time) ** self.shape,
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return np.asarray(cumulative_hazard_rate) ** (1 / self.shape) / self.rate

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_hf(self, time: AnyFloat) -> NumpyFloat:
        return np.stack(
            (
                self.rate
                * (self.rate * time) ** (self.shape - 1)
                * (1 + self.shape * np.log(self.rate * time)),
                self.shape**2 * (self.rate * time) ** (self.shape - 1),
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_chf(self, time: AnyFloat) -> NumpyFloat:
        return np.stack(
            (
                np.log(self.rate * time) * (self.rate * time) ** self.shape,
                self.shape * time * (self.rate * time) ** (self.shape - 1),
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def dhf(self, time: AnyFloat) -> NumpyFloat:
        time = np.asarray(time)
        return (
            self.shape
            * (self.shape - 1)
            * self.rate**2
            * (self.rate * time) ** (self.shape - 2)
        )


@final
class Gompertz(LifetimeDistribution):
    r"""
    Gompertz lifetime distribution.

    The Gompertz distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = c \lambda e^{\lambda t} e^{ -c \left( e^{\lambda t}-1 \right) }

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, default is None
        Shape parameter.
    rate : float, default is None
        Rate parameter.

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: float | None = None, rate: float | None = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the shape value.

        Returns
        -------
        out: float
        """
        return self._params["shape"]

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the rate value.

        Returns
        -------
        out: float
        """
        return self._params["rate"]

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def hf(self, time: AnyFloat) -> NumpyFloat:
        return self.shape * self.rate * np.exp(self.rate * time)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def chf(self, time: AnyFloat) -> NumpyFloat:
        return self.shape * np.expm1(self.rate * time)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def mean(self) -> NumpyFloat:
        return np.exp(self.shape) * exp1(self.shape) / self.rate

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def var(self) -> NumpyFloat:
        return super().var()

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def mrl(self, time: AnyFloat) -> NumpyFloat:
        z = self.shape * np.exp(self.rate * time)
        return np.exp(z) * exp1(z) / self.rate

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return 1 / self.rate * np.log1p(cumulative_hazard_rate / self.shape)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_hf(self, time: AnyFloat) -> NumpyFloat:
        return np.stack(
            (
                self.rate * np.exp(self.rate * time),
                self.shape * np.exp(self.rate * time) * (1 + self.rate * time),
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_chf(self, time: AnyFloat) -> NumpyFloat:
        return np.stack(
            (
                np.expm1(self.rate * time),
                self.shape * time * np.exp(self.rate * time),
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def dhf(self, time: AnyFloat) -> NumpyFloat:
        return self.shape * self.rate**2 * np.exp(self.rate * time)


@final
class Gamma(LifetimeDistribution):
    r"""
    Gamma lifetime distribution.

    The Gamma distribution is a 2-parameter distribution with
    :math:`(c,\lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{\lambda^c t^{c-1} e^{-\lambda t}}{\Gamma(c)}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, default is None
        Shape parameter.
    rate : float, default is None
        Rate parameter.

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: float | None = None, rate: float | None = None):
        super().__init__(shape=shape, rate=rate)

    def _uppergamma(self, x: AnyFloat) -> NumpyFloat:
        x = np.asarray(x, dtype=np.float64)
        return gammaincc(self.shape, x) * gamma(self.shape)

    def _jac_uppergamma_shape(self, x: AnyFloat) -> NumpyFloat:
        return laguerre_quadrature(
            lambda s: np.log(s) * s ** (self.shape - 1), x, deg=100
        )

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the shape value.

        Returns
        -------
        out: float
        """
        return self._params["shape"]

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the rate value.

        Returns
        -------
        out: float
        """
        return self._params["rate"]

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def hf(self, time: AnyFloat) -> NumpyFloat:
        x = np.asarray(self.rate * time, dtype=np.float64)
        return self.rate * x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def chf(self, time: AnyFloat) -> NumpyFloat:
        x = np.asarray(self.rate * time, dtype=np.float64)
        return np.log(gamma(self.shape)) - np.log(self._uppergamma(x))

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def mean(self) -> NumpyFloat:
        return np.float64(self.shape / self.rate)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def var(self) -> NumpyFloat:
        return np.float64(self.shape / (self.rate**2))

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return (
            1
            / self.rate
            * np.asarray(
                gammainccinv(self.shape, np.exp(-cumulative_hazard_rate)),
                dtype=np.float64,
            )
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_hf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * time
        y = x ** (self.shape - 1) * np.exp(-x) / self._uppergamma(x) ** 2
        jac = (
            y
            * (
                (self.rate * np.log(x) * self._uppergamma(x))
                - self.rate * self._jac_uppergamma_shape(x)
            ),
            y * ((self.shape - x) * self._uppergamma(x) + x**self.shape * np.exp(-x)),
        )
        return np.stack(jac)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_chf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * time
        jac = (
            digamma(self.shape) - self._jac_uppergamma_shape(x) / self._uppergamma(x),
            (x ** (self.shape - 1) * time * np.exp(-x) / self._uppergamma(x)),
        )
        return np.stack(jac)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def dhf(self, time: AnyFloat) -> NumpyFloat:
        return self.hf(time) * ((self.shape - 1) / time - self.rate + self.hf(time))

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def mrl(self, time: AnyFloat) -> NumpyFloat:
        return super().mrl(time)


@final
class LogLogistic(LifetimeDistribution):
    r"""
    Log-logistic probability distribution.

    The Log-logistic distribution is defined as a 2-parameter distribution
    :math:`(c, \lambda)`. The probability density function is:

    .. math::

        f(t) = \frac{c \lambda^c t^{c-1}}{(1+(\lambda t)^{c})^2}

    where:

        - :math:`c > 0`, the shape parameter,
        - :math:`\lambda > 0`, the rate parameter,
        - :math:`t\geq 0`, the operating time, age, cycles, etc.

    |

    Parameters
    ----------
    shape : float, default is None
        Shape parameter.
    rate : float, default is None
        Rate parameter.

    Attributes
    ----------
    fitting_results : FittingResults, default is None
        An object containing fitting results (AIC, BIC, etc.).
        If the model is not fitted, the value is None.
    nb_params
    params
    params_names
    plot
    shape
    rate
    """

    def __init__(self, shape: float | None = None, rate: float | None = None):
        super().__init__(shape=shape, rate=rate)

    @property
    def shape(self) -> float:  # optional but better for clarity and type checking
        """
        Returns the shape value.

        Returns
        -------
        out: float
        """
        return self._params["shape"]

    @property
    def rate(self) -> float:  # optional but better for clarity and type checking
        """Returns the rate value.

        Returns
        -------
        out: float
        """
        return self._params["rate"]

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def hf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * np.asarray(time)
        return self.shape * self.rate * x ** (self.shape - 1) / (1 + x**self.shape)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def chf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * time
        return np.log(1 + x**self.shape)

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def mean(self) -> NumpyFloat:
        b = np.pi / self.shape
        if self.shape <= 1:
            raise ValueError(
                f"Expectancy only defined for shape > 1: shape = {self.shape}"
            )
        return b / (self.rate * np.sin(b))

    @override
    @document_args(
        base_cls=FittableParametricLifetimeModel,
        args_docstring=[],
        returns=[docscrape.Parameter("out", "np.float64", [""])],
    )
    def var(self) -> NumpyFloat:
        b = np.pi / self.shape
        if self.shape <= 2:
            raise ValueError(
                f"Variance only defined for shape > 2: shape = {self.shape}"
            )
        return (1 / self.rate**2) * (2 * b / np.sin(2 * b) - b**2 / (np.sin(b) ** 2))

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def ichf(self, cumulative_hazard_rate: AnyFloat) -> NumpyFloat:
        return ((np.exp(cumulative_hazard_rate) - 1) ** (1 / self.shape)) / self.rate

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_hf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * time
        jac = (
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2)
            * (1 + x**self.shape + self.shape * np.log(self.rate * time)),
            (self.rate * x ** (self.shape - 1) / (1 + x**self.shape) ** 2)
            * (self.shape**2 / self.rate),
        )
        return np.stack(jac)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def jac_chf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * time
        jac = (
            (x**self.shape / (1 + x**self.shape)) * np.log(self.rate * time),
            (x**self.shape / (1 + x**self.shape)) * (self.shape / self.rate),
        )
        return np.stack(jac)

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def dhf(self, time: AnyFloat) -> NumpyFloat:
        x = self.rate * np.asarray(time)
        return (
            self.shape
            * self.rate**2
            * x ** (self.shape - 2)
            * (self.shape - 1 - x**self.shape)
            / (1 + x**self.shape) ** 2
        )

    @override
    @document_args(base_cls=LifetimeDistribution, args_docstring=[])
    def mrl(self, time: AnyFloat) -> NumpyFloat:
        return super().mrl(time)


Ts = TypeVarTuple("Ts")


@final
class EquilibriumDistribution(ParametricLifetimeModel[*Ts]):
    r"""
    Equilibrium distribution.

    The equilibirum distribution is the distribution that makes the renewal process
    stationnary.

    Parameters
    ----------
    baseline : any parametric lifetime model
        Lifetime model.

    References
    ----------
    .. [1] Ross, S. M. (1996). Stochastic stochastic_process. New York: Wiley.
    """

    baseline: ParametricLifetimeModel[*Ts]

    def __init__(self, baseline: ParametricLifetimeModel[*Ts]):
        super().__init__()
        self.baseline = baseline

    @override
    def cdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        return legendre_quadrature(
            lambda x: self.baseline.sf(x, *args), 0, time
        ) / self.baseline.mean(*args)

    @override
    def sf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        return 1 - self.cdf(time, *args)

    @override
    def pdf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        return self.baseline.sf(time, *args) / self.baseline.mean(*args)

    @override
    def hf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        return 1 / self.baseline.mrl(time, *args)

    @override
    def chf(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        return -np.log(self.sf(time, *args))

    @override
    def isf(self, probability: AnyFloat, *args: *Ts) -> NumpyFloat:
        def func(x: NDArray[np.float64]) -> np.float64:
            return np.sum(self.sf(x, *args) - probability)

        return newton(
            func,
            x0=np.asarray(self.baseline.isf(probability, *args)),
            args=args,
        )

    @override
    def ichf(
        self,
        cumulative_hazard_rate: AnyFloat,
        *args: *Ts,
    ) -> NumpyFloat:
        return self.isf(np.exp(-cumulative_hazard_rate), *args)


AnyInt: TypeAlias = int | np.int64 | NDArray[np.int64]


@final
class MinimumDistribution(FittableParametricLifetimeModel[*tuple[AnyInt, *Ts]]):
    r"""
    Series structure of n identical and independent components.

    The hazard function of the system is given by:

    .. math::

        h(t) = n \cdot  h_0(t)

    where :math:`h_0` is the baseline hazard function of the components.

    Parameters
    ----------
    baseline : lifetime distribution or regression
        Lifetime model.

    Examples
    --------

    Computing the survival (or reliability) function for 3 structures of 3,6 and
    9 identical and idependent components:

    .. code-block::

        model = MinimumDistribution(Weibull(2, 0.05))
        t = np.arange(0, 10, 0.1)
        n = np.array([3, 6, 9]).reshape(-1, 1)
        model.sf(t, n)
    """

    baseline: FittableParametricLifetimeModel[*Ts]

    def __init__(self, baseline: FittableParametricLifetimeModel[*Ts]):
        super().__init__()
        self.baseline = baseline

    @override
    def sf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return super().sf(time, *(n, *args))

    @override
    def pdf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return super().pdf(time, *(n, *args))

    @override
    def hf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return n * self.baseline.hf(time, *args)

    @override
    def chf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return n * self.baseline.chf(time, *args)

    @override
    def ichf(
        self,
        cumulative_hazard_rate: AnyFloat,
        n: AnyInt,
        *args: *Ts,
    ) -> NumpyFloat:
        return self.baseline.ichf(cumulative_hazard_rate / float(n), *args)

    @override
    def dhf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return n * self.baseline.dhf(time, *args)

    @override
    def jac_chf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return n * self.baseline.jac_chf(time, *args)

    @override
    def jac_hf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return n * self.baseline.jac_chf(time, *args)

    @override
    def jac_sf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        jac_chf, sf = (
            self.jac_chf(time, n, *args),
            self.sf(time, n, *args),
        )
        return -jac_chf * sf

    @override
    def jac_cdf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        return super().jac_cdf(time, n, *args)

    @override
    def jac_pdf(self, time: AnyFloat, n: AnyInt, *args: *Ts) -> NumpyFloat:
        jac_hf, hf = self.jac_hf(time, n, *args), self.hf(time, n, *args)
        jac_sf, sf = self.jac_sf(time, n, *args), self.sf(time, n, *args)
        return jac_hf * sf + jac_sf * hf

    @override
    def ls_integrate(
        self,
        func: Callable[[NDArray[np.float64]], NDArray[np.float64]],
        a: AnyFloat,
        b: AnyFloat,
        n: AnyInt,
        *args: *Ts,
        deg: int = 10,
    ) -> NumpyFloat:
        return super().ls_integrate(func, a, b, n, *args, deg=deg)

    @override
    def fit(
        self,
        time: NDArray[np.float64],
        args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        event: NDArray[np.bool_] | None = None,
        entry: NDArray[np.float64] | None = None,
        kwargs: MaximumLikelihoodOptimizerOptions | None = None,
    ) -> Self:
        if args is None:
            raise ValueError(
                "MinimumDistribution expects at least one additional argument in args"
            )
        return super().fit(
            time,
            args=args,
            event=event,
            entry=entry,
            kwargs=optimizer_options,
        )



from __future__ import annotations

import warnings
from typing import Any, Generic, Self, Sequence, TypeVarTuple

import numpy as np
from numpy.typing import NDArray

from relife.base import FittingResults, FrozenParametricModel, ParametricModel
from relife.data import NHPPData
from relife.lifetime_model._base import (
    FittableParametricLifetimeModel,
)
from relife.stochastic_process._sample import StochasticSampleMapping
from relife.typing import AnyFloat, NumpyFloat

Ts = TypeVarTuple("Ts")

__all__ = ["NonHomogeneousPoissonProcess", "FrozenNonHomogeneousPoissonProcess"]


class NonHomogeneousPoissonProcess(ParametricModel, Generic[*Ts]):
    """
    Non-homogeneous Poisson process.
    """

    lifetime_model: FittableParametricLifetimeModel[*Ts]
    fitting_results: FittingResults | None

    def __init__(self, lifetime_model: FittableParametricLifetimeModel[*Ts]):
        super().__init__()
        self.lifetime_model = lifetime_model
        self.fitting_results = None

    def intensity(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.lifetime_model.hf(time, *args)

    def cumulative_intensity(self, time: AnyFloat, *args: *Ts) -> NumpyFloat:
        """
        The cumulative intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self.lifetime_model.chf(time, *args)

    def freeze(self, *args: *Ts) -> FrozenNonHomogeneousPoissonProcess[*Ts]:
        """
        Freeze any arguments required by the process into the object data.

        Parameters
        ----------
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        FrozenParametricModel
        """
        return FrozenNonHomogeneousPoissonProcess(self, *args)

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], *args: *Ts, seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in a StochasticSampleMapping object.

        Parameters
        ----------
        nb_samples : int
            The number of samples
        time_window : tuple of two floats
            Time window in which data are sampled
        *args : float or np.ndarray
            Additional arguments needed by the model.
        seed : int, optional
            Random seed, by default None.

        """

        from relife.utils import get_model_nb_assets

        from ._sample import NonHomogeneousPoissonProcessIterable

        frozen_nhpp = self.freeze(*args)
        iterable = NonHomogeneousPoissonProcessIterable(
            frozen_nhpp, nb_samples, time_window=time_window, seed=seed
        )
        struct_array = np.concatenate(tuple(iterable))
        struct_array = np.sort(
            struct_array, order=("asset_id", "sample_id", "timeline")
        )
        return StochasticSampleMapping.from_struct_array(
            struct_array, get_model_nb_assets(frozen_nhpp), nb_samples
        )

    def generate_failure_data(
        self, nb_samples: int, time_window: tuple[float, float], *args: *Ts, seed=None
    ) -> dict[str, Any]:
        r"""
        .. warning:: Not implemented yet
        """
        raise NotImplementedError(
            "Failure data methods for stochastic processes will be introduced in a future release"
        )

        # from ._sample import NonHomogeneousPoissonProcessIterable

        # frozen_nhpp = self.freeze(*args)

        # iterable = NonHomogeneousPoissonProcessIterable(
        #     frozen_nhpp, nb_samples, time_window=time_window, seed=seed
        # )
        # struct_array = np.concatenate(tuple(iterable))
        # struct_array = np.sort(
        #     struct_array, order=("sample_id", "asset_id", "timeline")
        # )

        # first_ages_index = np.nonzero(struct_array["entry"] == time_window[0])
        # last_ages_index = np.nonzero(struct_array["age"] == time_window[1])

        # event_index = np.nonzero(struct_array["event"])

        # first_ages = struct_array[first_ages_index]["entry"].copy()
        # last_ages = struct_array[last_ages_index]["age"].copy()

        # assets_ids = np.char.add(
        #     np.char.add(
        #         np.full_like(
        #             struct_array[last_ages_index]["sample_id"], "S", dtype=np.str_
        #         ),
        #         struct_array[last_ages_index]["sample_id"].astype(np.str_),
        #     ),
        #     np.char.add(
        #         np.full_like(
        #             struct_array[last_ages_index]["asset_id"], "A", dtype=np.str_
        #         ),
        #         struct_array[last_ages_index]["asset_id"].astype(np.str_),
        #     ),
        # )

        # events_assets_ids = np.char.add(
        #     np.char.add(
        #         np.full_like(
        #             struct_array[event_index]["sample_id"], "S", dtype=np.str_
        #         ),
        #         struct_array[event_index]["sample_id"].astype(np.str_),
        #     ),
        #     np.char.add(
        #         np.full_like(struct_array[event_index]["asset_id"], "A", dtype=np.str_),
        #         struct_array[event_index]["asset_id"].astype(np.str_),
        #     ),
        # )
        # ages_at_events = struct_array[event_index]["age"].copy()

        # return {
        #     "ages_at_events": ages_at_events,
        #     "events_assets_ids": events_assets_ids,
        #     "first_ages": first_ages,
        #     "last_ages": last_ages,
        #     "assets_ids": assets_ids,
        # }

    def fit(
        self,
        ages_at_events: NDArray[np.float64],
        events_assets_ids: Sequence[str] | NDArray[np.int64],
        first_ages: NDArray[np.float64] | None = None,
        last_ages: NDArray[np.float64] | None = None,
        lifetime_model_args: NDArray[Any] | tuple[NDArray[Any], ...] | None = None,
        assets_ids: Sequence[str] | NDArray[np.int64] | None = None,
        **kwargs: Any,
    ) -> Self:
        """
        Estimation of the process parameters from recurrent failure data.

        Parameters
        ----------
        ages_at_events : 1D array of float
            Array of float containing the ages of each asset when the events occured
        events_assets_ids : sequence of hashable
            Sequence object containing the ids of each assets corresponding to each ages.
        first_ages : 1D array of float, optional
            Array of float containing the ages of each asset before observing events. If set, ``assets_ids`` is needed and its length
            must equal the size of ``first_ages``.
        last_ages : 1D array of float, optional
            Array of float containing the ages of each asset at the end of the observation period. If set, ``assets_ids`` is needed and its length
            must equal the size of ``last_ages``.
        lifetime_model_args : tuple of np.ndarray, optional
            Additional arguments needed by the model. If set, ``assets_ids`` is needed.
            For 1D array, the size must equal the length of ``assets_ids``. For 2D array (e.g. covar of regression),
            the length of first axis must equal the length of ``assets_ids``.
        assets_ids : sequence of hashable, optional
            Only needed if either ``first_ages``, ``last_ages`` or ``model_args`` is filled. It must be a sequence object
            containing the unique ids corresponding to each values contained in ``first_ages``, ``last_ages`` and/or ``model_args``

        Returns
        -------
        Self
            The current object with the estimated parameters setted inplace.

        Examples
        --------

        Ages of assets AB2 and CX13 at each event.

        >>> from relife.lifetime_model import Weibull
        >>> from relife.stochastic_process import NonHomogeneousPoissonProcess
        >>> nhpp = NonHomogeneousPoissonProcess(Weibull())
        >>> nhpp.fit(
            np.array([11., 13., 21., 25., 27.]),
            ("AB2", "CX13", "AB2", "AB2", "CX13"),
        )

        With additional information and model args (regression of 2 coefficients)

        >>> from relife.lifetime_model import ParametricProportionalHazard
        >>> nhpp = NonHomogeneousPoissonProcess(ParametricProportionalHazard())
        >>> nhpp.fit(
            np.array([11., 13., 21., 25., 27.]),
            ("AB2", "CX13", "AB2", "AB2", "CX13"),
            first_ages = np.array([10., 12.]),
            last_ages = np.array([35., 60.]),
            model_args = (np.array([[1.2, 5.5], [37.2, 22.2]]),) # 2d array of 2 raws (2 assets) and 2 columns (2 coefficients)
        )
        """
        warnings.warn(
            "Fit method of NHPP will change in a future release", DeprecationWarning
        )

        nhpp_data = NHPPData(
            ages_at_events,
            events_assets_ids,
            first_ages=first_ages,
            last_ages=last_ages,
            model_args=lifetime_model_args,
            assets_ids=assets_ids,
        )
        time, event, entry, args = nhpp_data.to_lifetime_data()
        optimizer = self.lifetime_model.init_likelihood(
            time, args, event, entry, **kwargs
        )
        fitting_results = optimizer.optimize()
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


class FrozenNonHomogeneousPoissonProcess(
    FrozenParametricModel[NonHomogeneousPoissonProcess[*Ts], *Ts]
):
    """
    Non-homogeneous Poisson process.
    """

    def intensity(self, time: AnyFloat) -> NumpyFloat:
        """
        The intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self._unfrozen_model.intensity(time, *self.args)

    def cumulative_intensity(self, time: AnyFloat) -> NumpyFloat:
        """
        The cumulative intensity function of the process.

        Parameters
        ----------
        time : float or np.ndarray
            Elapsed time value(s) at which to compute the function.
            If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
        *args : float or np.ndarray
            Additional arguments needed by the model.

        Returns
        -------
        np.float64 or np.ndarray
            Function values at each given time(s).
        """
        return self._unfrozen_model.cumulative_intensity(time, *self.args)

    def sample(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ) -> StochasticSampleMapping:
        """Renewal data sampling.

        This function will sample data and encapsulate them in an object.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        tf : float
            Time at the end of the observation.
        t0 : float, default 0
            Time at the beginning of the observation.
        seed : int, optional
            Random seed, by default None.

        """
        return self._unfrozen_model.sample(
            nb_samples, time_window, *self.args, seed=seed
        )

    def generate_failure_data(
        self, nb_samples: int, time_window: tuple[float, float], seed=None
    ):
        """Generate failure data

        This function will generate failure data that can be used to fit a non-homogeneous Poisson process.

        Parameters
        ----------
        nb_samples : int
            The number of samples.
        time_window : tuple of two floats
            Time window in which data are sampled
        seed : int, optional
            Random seed, by default None.

        Returns
        -------
        A dict of ages_at_events, events_assets_ids, first_ages, last_ages, model_args and assets_ids
        """
        return self._unfrozen_model.generate_failure_data(
            nb_samples, time_window, *self.args, seed=seed
        )


def is_non_homogeneous_poisson_process(model):
    """
    Checks if model is a non-homogeneous Poisson process.
    """
    # local import to avoid circular import

    return isinstance(
        model, (NonHomogeneousPoissonProcess, FrozenNonHomogeneousPoissonProcess)
    )



# pyright: basic
import numpy as np
import pytest
from numpy.typing import NDArray

from relife.data import load_insulator_string, load_power_transformer
from relife.lifetime_model import (
    AgeReplacementModel,
    Exponential,
    Gamma,
    Gompertz,
    LogLogistic,
    ParametricAcceleratedFailureTime,
    ParametricProportionalHazard,
    Weibull,
)

#######################################################################################
# DATA FIXTURES
#######################################################################################


@pytest.fixture
def power_transformer_data():
    return load_power_transformer()


@pytest.fixture
def insulator_string_data():
    return load_insulator_string()


#######################################################################################
# LIFETIME MODEL FIXTURES
#######################################################################################


def exponential():
    return Exponential(0.00795203)


def weibull():
    return Weibull(3.46597395, 0.01227849)


def gompertz():
    return Gompertz(0.00865741, 0.06062632)


def gamma():
    return Gamma(5.3571091, 0.06622822)


def loglogistic():
    return LogLogistic(3.92614064, 0.0133325)


COEFFICIENTS = (np.log(2), np.log(2))


def pph_exponential():
    return ParametricProportionalHazard(exponential(), coefficients=COEFFICIENTS)


def pph_weibull():
    return ParametricProportionalHazard(weibull(), coefficients=COEFFICIENTS)


def pph_gompertz():
    return ParametricProportionalHazard(gompertz(), coefficients=COEFFICIENTS)


def pph_gamma():
    return ParametricProportionalHazard(gamma(), coefficients=COEFFICIENTS)


def pph_loglogistic():
    return ParametricProportionalHazard(loglogistic(), coefficients=COEFFICIENTS)


def aft_exponential():
    return ParametricAcceleratedFailureTime(exponential(), coefficients=COEFFICIENTS)


def aft_weibull():
    return ParametricAcceleratedFailureTime(weibull(), coefficients=COEFFICIENTS)


def aft_gompertz():
    return ParametricAcceleratedFailureTime(gompertz(), coefficients=COEFFICIENTS)


def aft_gamma():
    return ParametricAcceleratedFailureTime(gamma(), coefficients=COEFFICIENTS)


def aft_loglogistic():
    return ParametricAcceleratedFailureTime(loglogistic(), coefficients=COEFFICIENTS)


@pytest.fixture(params=[exponential(), weibull(), gompertz(), gamma(), loglogistic()])
def distribution(request):
    yield request.param


@pytest.fixture(
    params=[
        pph_exponential(),
        pph_weibull(),
        pph_gompertz(),
        pph_gamma(),
        pph_loglogistic(),
        aft_exponential(),
        aft_weibull(),
        aft_gompertz(),
        aft_gamma(),
        aft_loglogistic(),
    ]
)
def regression(request):
    yield request.param


#######################################################################################
# LIFETIME LIKELIHOOD FIXTURES
#######################################################################################


@pytest.fixture
def distribution_likelihood(distribution, power_transformer_data):
    return distribution.init_likelihood(
        power_transformer_data["time"],
        event=power_transformer_data["event"],
        entry=power_transformer_data["entry"],
    )


@pytest.fixture
def regression_likelihood(regression, insulator_string_data):
    covar = np.column_stack(
        (
            insulator_string_data["pHCl"],
            insulator_string_data["pH2SO4"],
        )
    )
    return regression.init_likelihood(
        insulator_string_data["time"],
        covar,
        event=insulator_string_data["event"],
        entry=insulator_string_data["entry"],
    )


#######################################################################################
# FROZEN LIFETIME FIXTURES
#######################################################################################

NB_ASSETS = 3


@pytest.fixture
def frozen_regression(regression):
    covar = np.linspace(0.0, 0.5, num=NB_ASSETS * regression.nb_coef).reshape(
        NB_ASSETS, regression.nb_coef
    )
    return regression.freeze(covar)


@pytest.fixture
def frozen_ar_distribution(distribution):
    ar = distribution.isf(0.75)
    return AgeReplacementModel(distribution).freeze(ar)


@pytest.fixture
def frozen_ar_regression(regression):
    covar = np.linspace(0.0, 0.5, num=NB_ASSETS * regression.nb_coef).reshape(
        NB_ASSETS, regression.nb_coef
    )
    ar = regression.isf(0.75, covar)
    return AgeReplacementModel(regression).freeze(ar, covar)


#######################################################################################
# LIFETIME MODEL VARIABLES FIXTURES
#######################################################################################

M = 3  # nb assets
N = 10  # nb points


@pytest.fixture(
    params=[
        np.float64(1),
        np.ones((1,), dtype=np.float64),
        np.ones((N,), dtype=np.float64),
        np.ones((1, 1), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
        np.ones((1, N), dtype=np.float64),
        np.ones((M, 3), dtype=np.float64),
    ],
    ids=lambda time: f"time:{time.shape}",
)
def time(request):
    return request.param


@pytest.fixture(
    params=[
        np.float64(0.5),
        np.ones((1,), dtype=np.float64) * 0.5,
        np.ones((N,), dtype=np.float64) * 0.5,
        np.ones((1, 1), dtype=np.float64) * 0.5,
        np.ones((M, 1), dtype=np.float64) * 0.5,
        np.ones((1, N), dtype=np.float64) * 0.5,
        np.ones((M, N), dtype=np.float64) * 0.5,
    ],
    ids=lambda probability: f"probability:{probability.shape}",
)
def probability(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((len(COEFFICIENTS),), dtype=np.float64),
        np.ones((1, len(COEFFICIENTS)), dtype=np.float64),
        np.ones((M, len(COEFFICIENTS)), dtype=np.float64),
    ],
    ids=lambda covar: f"covar:{covar.shape}",
)
def covar(request):
    return request.param


@pytest.fixture(
    params=[
        2.0 * np.ones((), dtype=np.float64),
        2.0 * np.ones((1,), dtype=np.float64),
        2.0 * np.ones((N,), dtype=np.float64),
        2.0 * np.ones((1, N), dtype=np.float64),
        2.0 * np.ones((M, 1), dtype=np.float64),
        2.0 * np.ones((M, N), dtype=np.float64),
    ],
    ids=lambda a: f"a:{a.shape}",
)
def integration_bound_a(request):
    return request.param


@pytest.fixture(
    params=[
        8.0 * np.ones((), dtype=np.float64),
        8.0 * np.ones((1,), dtype=np.float64),
        8.0 * np.ones((N,), dtype=np.float64),
        8.0 * np.ones((1, N), dtype=np.float64),
        8.0 * np.ones((M, 1), dtype=np.float64),
        8.0 * np.ones((M, N), dtype=np.float64),
    ],
    ids=lambda b: f"b:{b.shape}",
)
def integration_bound_b(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda ar: f"ar:{ar.shape}",
)
def ar(request):
    return request.param


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda a0: f"a0:{a0.shape}",
)
def a0(request):
    return request.param


@pytest.fixture(
    params=[1, N],
    ids=lambda size: f"size:{size}",
)
def rvs_size(request):
    return request.param


#######################################################################################
# ECONOMIC FIXTURES
#######################################################################################


@pytest.fixture(
    params=[
        np.ones((), dtype=np.float64),
        # np.ones((1,), dtype=np.float64),
        np.ones((M,), dtype=np.float64),
        # np.ones((M, 1), dtype=np.float64),
    ],
    ids=lambda cp: f"cp:{cp.shape}",
)
def cp(request):
    return request.param


# M = 3
CF_RANGE = [5, 10, 20, 100, 1000]


@pytest.fixture(
    params=[
        np.array(CF_RANGE[0], dtype=np.float64),  # ()
        # np.array([CF_RANGE[0]], dtype=np.float64), # (1,)
        np.array(CF_RANGE[:M], dtype=np.float64),  # (M,)
        # np.array(CF_RANGE[:M], dtype=np.float64).reshape(-1, 1), # (M, 1)
    ],
    ids=lambda cf: f"cf:{cf.shape}",
)
def cf(request):
    return request.param


@pytest.fixture(params=[0.0, 0.04], ids=lambda rate: f"discounting_rate:{rate}")
def discounting_rate(request):
    return request.param


#######################################################################################
# FACTORY FIXTURE TO CONTROL IO SHAPES
#######################################################################################


@pytest.fixture
def expected_out_shape():
    def _expected_out_shape(**kwargs: NDArray[np.float64]) -> tuple[int, ...]:
        def shape_contrib(**kwargs: NDArray[np.float64]):
            yield ()  # yield at least (), in case kwargs is empty
            for k, v in kwargs.items():
                match k:
                    case "covar" if v.ndim == 2:
                        yield v.shape[0], 1
                    case "covar" if v.ndim < 2:
                        yield ()
                    case "cf" | "cp" | "ar" | "a0" if v.ndim == 2 or v.ndim == 0:
                        yield v.shape
                    case "cf" | "cp" | "ar" | "a0" if v.ndim == 1:
                        yield v.size, 1
                    case "size":
                        if isinstance(v, int):
                            if v == 1:
                                yield ()
                            yield (v,)
                        yield v  # it is tuple
                    case _:
                        yield v.shape

        return np.broadcast_shapes(*tuple(shape_contrib(**kwargs)))

    return _expected_out_shape



# pyright: basic

from __future__ import annotations

from collections.abc import Mapping
from typing import Iterator, Sequence, TypedDict

import numpy as np
from numpy.typing import NDArray

__all__ = ["StochasticSampleMapping"]


class StochasticSample(TypedDict):
    """
    Container for 2D matrixes of stochastic samples.
    Assets x Samples on axis 0 and Timeline on axis 1.
    Values are either booleans to indicate events and renewals, or floats for rewards.
    """

    events: NDArray[np.bool_]
    preventive_renewals: NDArray[np.bool_]
    rewards: NDArray[np.float64] | None


class StochasticSampleMapping(
    Mapping[str, NDArray[np.float64] | NDArray[np.bool_] | None]
):
    """
    Mapping class to manipulate Stochastic samples.
    Access to 2D matrixes of events, renewals and rewards and method for selecting sub-samples of the full data.
    """

    nb_assets: int
    nb_samples: int
    timeline: NDArray[np.float64]

    _stochastic_data_sample: StochasticSample

    def __init__(
        self,
        nb_assets: int,
        nb_samples: int,
        timeline: NDArray[np.float64],
        stochastic_data_sample: StochasticSample,
    ) -> None:
        self.nb_assets = nb_assets
        self.nb_samples = nb_samples
        self.timeline = timeline
        self._stochastic_data_sample = stochastic_data_sample

    @classmethod
    def from_struct_array(
        cls, struct_array: NDArray[np.void], nb_assets: int, nb_samples: int
    ) -> StochasticSampleMapping:
        """
        Class method to build the _StochasticSample and timeline from a struct_array before initiating.
        """

        # assets x samples are placed on axis 0
        index_in_rows = (
            struct_array["asset_id"] * nb_samples + struct_array["sample_id"]
        )
        _, row_index = np.unique(index_in_rows, return_inverse=True)

        # unique values of timeline on axis 1
        timeline, col_timeline = np.unique(
            struct_array["timeline"], return_inverse=True
        )

        # construction of matrixes
        events = np.zeros((nb_assets * nb_samples, timeline.size), dtype=bool)
        events[row_index, col_timeline] = struct_array["event"]
        preventive_renewals = np.zeros(
            (nb_assets * nb_samples, timeline.size), dtype=bool
        )
        preventive_renewals[row_index, col_timeline] = ~struct_array["event"]
        preventive_renewals[:, -1] = False

        if "rewards" in struct_array.dtype.fields.keys():
            rewards = np.zeros((nb_assets * nb_samples, timeline.size), dtype=float)
            rewards[row_index, col_timeline] = struct_array["reward"]
        else:
            rewards = None

        stochastic_data_sample = {
            "events": events,
            "preventive_renewals": preventive_renewals,
            "rewards": rewards,
        }

        return cls(
            nb_assets=nb_assets,
            nb_samples=nb_samples,
            timeline=timeline,
            stochastic_data_sample=stochastic_data_sample,
        )

    def __getitem__(self, key: str) -> NDArray[np.float64] | NDArray[np.bool_] | None:
        if key not in self._stochastic_data_sample:
            raise KeyError(
                f"Key {key} does not exists. Allowed keys are 'events', 'preventive_renewals' or 'rewards'"
            )
        return self._stochastic_data_sample.get(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self._stochastic_data_sample)

    def __len__(self) -> int:
        return len(self._stochastic_data_sample)

    def select(
        self,
        sample_id: int | Sequence[int] | None = None,
        asset_id: int | Sequence[int] | None = None,
    ) -> StochasticSampleMapping:
        """
        Select a sub part of the sample based on assets and samples ids.
        Returns a truncated StochasticSampleMapping by selecting the corresponding rows in the 2D matrixes.
        """
        if asset_id is None:
            asset_id = np.arange(self.nb_assets)
        if sample_id is None:
            sample_id = np.arange(self.nb_samples)

        asset_id = np.atleast_1d(asset_id)
        sample_id = np.atleast_1d(sample_id)

        new_nb_assets = asset_id.shape[0]
        new_nb_samples = sample_id.shape[0]

        mask = (asset_id[None, :] * self.nb_samples + sample_id[:, None]).flatten()

        new_stochastic_data_sample = {
            key: value[mask] for key, value in self._stochastic_data_sample.items()
        }

        return StochasticSampleMapping(
            nb_assets=new_nb_assets,
            nb_samples=new_nb_samples,
            timeline=self.timeline,
            stochastic_data_sample=new_stochastic_data_sample,
        )
